;;;; MaiSQL --- Common Lisp Interface Layer to SQL Databases
;;;; This is copyrighted software.  See documentation for terms.
;;;; 
;;;; postgresql-sql.cl --- SQL-Interface implementation for PostgreSQL
;;;; 
;;;; Checkout Tag: $Name: debian_version_1_2_1-1 $
;;;; $Id: postgresql-sql.lisp,v 1.41 2003/03/26 17:11:21 craig Exp $

(in-package :MAISQL-POSTGRESQL)

;;;; %File Description:
;;;; 
;;;; 
;;;; 

(defvar *max-identifier-length* 32)

(defmethod database-initialize-database-type
  ((database-type (eql :postgresql)))
  t)

(defmethod database-name-from-spec
  (connection-spec (database-type (eql :postgresql)))
  (check-connection-spec connection-spec database-type
                         (host db user password &optional port options tty))
  (destructuring-bind (host db user password &optional port options tty)
      connection-spec
    (declare (ignore password options tty))
    (concatenate 'string host (if port ":") (if port port) "/" db "/" user)))

(defmethod database-busy ((database postgresql-database))
  (pqisbusy (database-conn-ptr database)))

(defmethod database-connect (connection-spec (database-type (eql :postgresql)))
  (check-connection-spec connection-spec database-type
			 (host db user password &optional port options tty))
  (let ((connection (%pg-database-connection connection-spec)))
    ;; Success, make instance
    (make-instance 'postgresql-database
                   :name (database-name-from-spec connection-spec
                                                  database-type)
                   :spec connection-spec
                   :db-type :postgresql
                   :conn-ptr connection)))

(defmethod database-disconnect ((database postgresql-database))
    (let ((foo (database-lock database)))
      (with-lock (foo)
	(pqfinish (database-conn-ptr database))
	(setf (database-conn-ptr database) nil)
	t)))

(defun %trim-crlf (string)
  (string-right-trim '(#\Return #\Newline) string))

(defmethod database-query (query-expression (database postgresql-database))
  (flet ((error-text ()
           (%trim-crlf (pqerrormessage (database-conn-ptr database)))))
    (with-lock ((database-lock database))
      (let ((result (%pg-query database query-expression)))
        (unwind-protect
             (case (pqresultstatus result)
               (:empty-query
                nil)
               (:tuples-ok
                (loop for tuple-index from 0 below (pqntuples result)
                      collect
                      (loop for i from 0 below (pqnfields result)
                            collect
                            (if (pqgetisnull result tuple-index i)
                                nil
                                (pqgetvalue result tuple-index i)))))
               (:fatal-error
                (let ((msg-text (subseq (error-text) 8)))
                  (error 'maisql-sql-error
                         :database database
                         :expression query-expression
                         :errno (pqresultstatus result)
                         :error msg-text)))
               (t
                (error 'maisql-sql-error
                       :database database
                       :expression query-expression
                       :errno (pqresultstatus result)
                       :error (concatenate 'string
                                           (error-text)
                                           (pqcmdstatus result)))))
          (pqclear result))))))

(defmethod database-create-sequence
  (sequence-name (database postgresql-database))
  (let ((tref (sql-escape (etypecase sequence-name
                            (string (string-upcase sequence-name))
                            (sql-sys::sql-ident
                             (sql sequence-name))
                            (symbol
                             (sql sequence-name))))))
    (database-execute-command
     (concatenate 'string "CREATE SEQUENCE " tref) database)))
  
(defmethod database-drop-sequence
  (sequence-name (database postgresql-database))
    (let ((tref (sql-escape (etypecase sequence-name
                              (string (string-upcase sequence-name))
                              (sql-sys::sql-ident
                               (sql sequence-name))
                              (symbol
                               (sql sequence-name))))))
      (database-execute-command
       (concatenate 'string "DROP SEQUENCE " tref) database)))

(defmethod database-sequence-next (sequence-name (database postgresql-database))
  (let ((tref (sql-escape (etypecase sequence-name
                            (string (string-upcase sequence-name))
                            (sql-sys::sql-ident
                             (sql sequence-name))
                            (symbol
                             (sql sequence-name))))))
    (parse-integer
     (caar
      (database-query
       (concatenate 'string "SELECT NEXTVAL ('" tref "')")
       database)))))

(defmethod database-sequence-last (sequence-name (database postgresql-database))
  (let ((tref (sql-escape (etypecase sequence-name
                            (string (string-upcase sequence-name))
                            (sql-sys::sql-ident
                             (sql sequence-name))
                            (symbol
                             (sql sequence-name))))))
    (parse-integer
     (caar
      (database-query
       (concatenate 'string "SELECT LAST_VALUE ('" tref "')")
       database)))))

(defmethod database-set-sequence-position (sequence-name (position integer)
							 (database postgresql-database))
  (let ((tref (sql-escape (etypecase sequence-name
                            (string (string-upcase sequence-name))
                            (sql-sys::sql-ident
                             (sql sequence-name))
                            (symbol
                             (sql sequence-name))))))
    (parse-integer
     (caar
      (database-query
       (format nil "SELECT SETVAL ('~A', ~A)" tref position)
       ;(concatenate 'string "SELECT SETVAL ('" tref "', " position ")")
       database)))))

(defmethod database-execute-command
    (sql-expression (database postgresql-database))
    (with-lock ((database-lock database))
      (let ((conn-ptr (database-conn-ptr database))
	    (result (%pg-exec database sql-expression)))
      (unwind-protect
           (case (pqresultstatus result)
             (:command-ok
              t)
             (:empty-query t)
             ;; (warn ";; Warning, database-execute-command:\
             ;; Empty statement sent to server."))
             (:tuples-ok t)
             ;; (warn ";; Warning, database-execute-command:\
             ;; Recieved tuples ok in response to pqexec."))
             (t
              (error 'maisql-sql-error
                     :database database
                     :expression sql-expression
                     :errno (pqresultstatus result)
                     :error (concatenate 'string
                                         (%trim-crlf (pqerrormessage conn-ptr))
                                         (pqcmdstatus result)))))
        (pqclear result)))))

(defstruct postgresql-result-set
  #+cmu (res-ptr (sap-alien (int-sap 0) psql-result)
                 :type (alien psql-result))
  #+lispworks (res-ptr NIL)
  (num-tuples 0)
  (num-fields 0)
  (tuple-index 0))

(defmethod database-query-result-set
  (query-expression (database postgresql-database) &optional full-set)
  (let ((result (%pg-query-result-set database query-expression)))
    (case (pqresultstatus result)
      ((:empty-query :tuples-ok)
       (if full-set
           (values (make-postgresql-result-set
                    :res-ptr result
                    :num-fields (pqnfields result)
                    :num-tuples (pqntuples result))
                   (pqnfields result)
                   (pqntuples result))
           (values (make-postgresql-result-set
                    :res-ptr result
                    :num-fields (pqnfields result)
                    :num-tuples (pqntuples result))
                   (pqnfields result))))
      (t
       (unwind-protect
            (error 'maisql-sql-error
                   :database database
                   :expression query-expression
                   :errno (pqresultstatus result)
                   :error (pqcmdstatus result))
         (pqclear result))))))

(defmethod database-dump-result-set (result-set (database postgresql-database))
    (let ((res-ptr (postgresql-result-set-res-ptr result-set)))
      (declare (type psql-result-ptr res-ptr))
      (pqclear res-ptr)
      t))

(defmethod database-store-next-row
  (result-set (database postgresql-database) list)
  (let ((result (postgresql-result-set-res-ptr result-set)))
    (declare (type psql-result-ptr result))
    (if (>= (postgresql-result-set-tuple-index result-set)
	    (postgresql-result-set-num-tuples result-set))
	nil
	(loop with tuple-index = (postgresql-result-set-tuple-index result-set)
	      for i from 0 below (postgresql-result-set-num-fields result-set)
	      for rest on list
	      do
	      (setf (car rest)
		    (if (pqgetisnull result tuple-index i)
			nil
			(pqgetvalue result tuple-index i)))
	      finally
	      (incf (postgresql-result-set-tuple-index result-set))
	      (return list)))))


(defmethod database-list-tables ((database postgresql-database)
				 &key (system-tables nil))
  (let ((res (select [tablename] :from [pg_tables] :flatp t
                     :database database)))
    (if (not system-tables)
        (remove-if #'(lambda (table)
		       (equal (subseq table 0 3)
			      "pg_")) res)
        res)))

(defmethod database-list-sequences ((database postgresql-database)
                                    &key (system-sequences nil))
  (let ((res (select [relname] :from [pg_class] :flatp t
                     :where [= [relkind] "S"]
                     :database database)))
    (if (not system-sequences)
	(remove-if #'(lambda (table)
		       (equal (subseq table 0 3)
			      "pg_")) res)
        res)))


(defmethod database-indexp ((database postgresql-database) name)
  (let ((tref (sql-escape (etypecase name
                            (string (string-upcase name))
                            (sql-sys::sql-ident
                             (sql name))
                            (symbol
                             (sql name))))))
    (when (and tref
               (member (string-upcase (subseq tref 0 (min (length tref) (1- *max-identifier-length*))))
                       (mapcar #'string-upcase
                               (database-list-indexes  database :system-indexes t))
                       :test #'equal))
      t)))

(defmethod database-sequencep ((database postgresql-database) name)
  (let ((tref (etypecase name
                (string (string-upcase name))
                (sql-sys::sql-ident
                 (sql name))
                (symbol
                 (sql-escape (sql name))))))
    (when (and tref
               (member (string-upcase (subseq tref 0 (min (length tref) (1- *max-identifier-length*))))
                       (mapcar #'string-upcase
                               (database-list-sequences database :system-sequences t))
                       :test #'equal))
      t)))

(defmethod database-tablep ((database postgresql-database) name)
  (let ((tref (etypecase name
                (string (string-upcase name))
                (sql-sys::sql-ident
                 (sql name))
                (symbol
                 (sql-escape (sql name))))))
    (when (and tref
               (member (string-upcase (subseq tref 0 (min (length tref) (1- *max-identifier-length*))))
                       (mapcar #'string-upcase
                               (database-list-tables database :system-tables t))
                       :test #'equal))
      t)))

(defmethod database-list-indexes ((database postgresql-database)
                                    &key (system-indexes nil))
  (let ((res (select [relname] :from [pg_class] :flatp t
                     :where [= [relkind] "i"]
                     :database database)))
    (if (not system-indexes)
	(remove-if #'(lambda (table)
		       (equal (subseq table 0 3)
			      "pg_")) res)
        res)))

(defmethod database-list-attributes (table (database postgresql-database))
  (let* ((relname (etypecase table
                     (sql-sys::sql-ident
                      (sql-escape
                       (string-downcase
                        (symbol-name (slot-value table 'sql-sys::name)))))
                     (string (sql-escape table))))
	 (result (select [pg_attribute attname]
			 :from '([pg_class] [pg_attribute])
			 :where [and [= [pg_class oid] [pg_attribute attrelid]]
				     [= [pg_class relname] relname]] :flatp t)))
    (if result
	(reverse
         (remove-if #'(lambda (it) (member it '("cmin"
                                                "cmax"
                                                "xmax"
                                                "xmin"
					       "oid"
                                                "ctid") :test #'equal)) result)))))

(defmethod database-attribute-type (attribute table
                                              (database postgresql-database))
  (let* ((relname (etypecase table
		    (sql-sys::sql-ident
		     (sql-escape
                      (string-downcase
                       (symbol-name (slot-value table 'sql-sys::name)))))
		    (string (sql-escape table))))
         (attname (etypecase attribute
		    (sql-sys::sql-ident
		     (sql-escape
                      (string-downcase
                       (symbol-name (slot-value attribute 'sql-sys::name)))))
		    (string (sql-escape attribute))))
         (result
          (select [pg_type typname] 
                  :from '([pg_type] [pg_class] [pg_attribute])
                  :where [and [= [pg_class oid] [pg_attribute attrelid]]
                              [= [pg_class relname] relname]
                              [= [pg_attribute attname] attname]
                              [= [pg_attribute atttypid] [pg_type oid]]]
                 :flatp t)))
    (if result
	(intern (string-upcase (car result)) :keyword) nil)))


;;(defmethod database-add-attribute (table attribute database)
;;  (let ((attname (slot-value attribute 'name)


(defmethod database-output-sql ((expr sql-sys::sql-typecast-exp) (database postgresql-database))
  (with-slots (sql-sys::modifier sql-sys::components)
    expr
    (if sql-sys::modifier
        (progn
          (sql-sys::output-sql sql-sys::components database)
          (write-char #\: sql-sys::*sql-stream*)
          (write-char #\: sql-sys::*sql-stream*)
          (write-string (symbol-name sql-sys::modifier) sql-sys::*sql-stream*)))))

(defmethod database-output-sql-as-type ((type (eql 'integer)) val (database postgresql-database))
  ;; typecast it so it uses the indexes
  (when val
    (make-instance 'sql-sys::sql-typecast-exp
                   :modifier 'int8
                   :components val)))

;; fix db version specific types
(defmethod %%database-get-version ((database postgresql-database))
  (odcl::when-bind (version (caar (database-query
			     (format nil "SELECT VERSION()")
			     database)))
	     (odcl::split (second (odcl::split version #\space)) #\.)))

(defmethod maisql-sys::database-get-type-specifier ((type (eql 'odcl::wall-time)) args (database postgresql-database))
  (declare (ignore args))
  (if (string= (car (%%database-get-version database)) "7")
      "TIMESTAMP WITHOUT TIME ZONE"
      "DATETIME"))

(defun pg6-timestring (time)
  "return the string to store the given time in the database"
  (declare (optimize (speed 3)))
  (let ((output (copy-seq "'XXXX-XX-XX XX:XX:XX-00'")))
    (flet ((inscribe-base-10 (output offset size decimal)
             (declare (type fixnum offset size decimal)
                      (type (simple-vector 10) odcl::+decimal-printer+))
             (dotimes (x size)
               (declare (type fixnum x)
                        (optimize (safety 0)))
               (multiple-value-bind (next this)
                   (floor decimal 10)
                 (setf (aref output (+ (- size x 1) offset))
                       (aref odcl::+decimal-printer+ this))
                 (setf decimal next)))))
      (multiple-value-bind (second minute hour day month year)
          (odcl::decode-time time)
        (inscribe-base-10 output 1 4 year)
        (inscribe-base-10 output 6 2 month)
        (inscribe-base-10 output 9 2 day)
        (inscribe-base-10 output 12 2 hour)
        (inscribe-base-10 output 15 2 minute)
        (inscribe-base-10 output 18 2 second)
        output))))

(defmethod sql-sys::database-output-sql ((self odcl::wall-time) database)
  (if (string= (car (%%database-get-version database)) "7")
      (odcl::db-timestring self)
      (pg6-timestring self)))


